(function (global) {

  var hasLocalStorage;

  // wrapper to implement defer() functionality
  Promise.defer = function () {
    var resolve, reject;
    var promise = new Promise(function() {
      resolve = arguments[0];
      reject = arguments[1];
    });
    return {
        resolve: resolve,
      reject: reject,
      promise: promise
    };
  };

  function logError(error) {
    if (typeof(error) === 'string') {
      console.error(error);
    } else {
      console.error(error.message, error.stack);
    }
  }

  function emitUnauthorized(r) {
    if (r.statusCode === 403  || r.statusCode === 401) {
      this._emit('error', new RemoteStorage.Unauthorized());
    }
    return Promise.resolve(r);
  }

  function shareFirst(path) {
    return ( this.backend === 'dropbox' &&
             path.match(/^\/public\/.*[^\/]$/) );
  }

  var SyncedGetPutDelete = {
    get: function (path, maxAge) {
      var self = this;
      if (this.local) {
        if (maxAge === undefined) {
          if ((typeof this.remote === 'object') &&
               this.remote.connected && this.remote.online) {
            maxAge = 2*this.getSyncInterval();
          } else {
            RemoteStorage.log('Not setting default maxAge, because remote is offline or not connected');
            maxAge = false;
          }
        }
        var maxAgeInvalid = function (maxAge) {
          return maxAge !== false && typeof(maxAge) !== 'number';
        };

        if (maxAgeInvalid(maxAge)) {
          return Promise.reject('Argument \'maxAge\' must be false or a number');
        }
        return this.local.get(path, maxAge, this.sync.queueGetRequest.bind(this.sync));
      } else {
        return this.remote.get(path);
      }
    },

    put: function (path, body, contentType) {
      if (shareFirst.bind(this)(path)) {
        return SyncedGetPutDelete._wrapBusyDone.call(this, this.remote.put(path, body, contentType));
      }
      else if (this.local) {
        return this.local.put(path, body, contentType);
      } else {
        return SyncedGetPutDelete._wrapBusyDone.call(this, this.remote.put(path, body, contentType));
      }
    },

    'delete': function (path) {
      if (this.local) {
        return this.local.delete(path);
      } else {
        return SyncedGetPutDelete._wrapBusyDone.call(this, this.remote.delete(path));
      }
    },

    _wrapBusyDone: function (result) {
      var self = this;
      this._emit('wire-busy');
      return result.then(function (r) {
        self._emit('wire-done', { success: true });
        return Promise.resolve(r);
      }, function (err) {
        self._emit('wire-done', { success: false });
        return Promise.reject(err);
      });
    }
  };

  /**
   * Class: RemoteStorage
   *
   * TODO needs proper introduction and links to relevant classes etc
   *
   * Constructor for global remoteStorage object.
   *
   * This class primarily contains feature detection code and a global convenience API.
   *
   * Depending on which features are built in, it contains different attributes and
   * functions. See the individual features for more information.
   *
   *  (start code)
   *  var remoteStorage = new RemoteStorage({
   *    logging: true,  // defaults to false
   *    cordovaRedirectUri: 'https://app.mygreatapp.com' // defaults to undefined
   *  });
   *  (end code)
   */
  var RemoteStorage = function (cfg) {
    /**
     * Event: ready
     *
     * Fired when ready
     **/
    /**
     * Event: not-connected
     *
     * Fired when ready, but no storage connected ("anonymous mode")
     **/
    /**
     * Event: connected
     *
     * Fired when a remote storage has been connected
     **/
    /**
     * Event: disconnected
     *
     * Fired after disconnect
     **/
    /**
     * Event: error
     *
     * Fired when an error occurs
     *
     * Arguments:
     * the error
     **/
    /**
     * Event: features-loaded
     *
     * Fired when all features are loaded
     **/
    /**
     * Event: connecting
     *
     * Fired before webfinger lookup
     **/
    /**
     * Event: authing
     *
     * Fired before redirecting to the authing server
     **/
    /**
     * Event: wire-busy
     *
     * Fired when a wire request starts
     **/
    /**
     * Event: wire-done
     *
     * Fired when a wire request completes
     **/
    /**
     * Event: network-offline
     *
     * Fired once when a wire request fails for the first time, and
     * `remote.online` is set to false
     **/
    /**
     * Event: network-online
     *
     * Fired once when a wire request succeeds for the first time after a
     * failed one, and `remote.online` is set back to true
     **/

    // Initial configuration property settings.
    if (typeof cfg === 'object') {
      RemoteStorage.config.logging = !!cfg.logging;
      RemoteStorage.config.cordovaRedirectUri = cfg.cordovaRedirectUri;
    }

    RemoteStorage.eventHandling(
      this, 'ready', 'connected', 'disconnected', 'not-connected', 'conflict',
            'error', 'features-loaded', 'connecting', 'authing',
            'sync-interval-change', 'wire-busy', 'wire-done',
            'network-offline', 'network-online'
    );

    // pending get/put/delete calls.
    this._pending = [];

    this._setGPD({
      get: this._pendingGPD('get'),
      put: this._pendingGPD('put'),
      delete: this._pendingGPD('delete')
    });

    this._cleanups = [];

    this._pathHandlers = { change: {} };

    this.apiKeys = {};

    hasLocalStorage = RemoteStorage.util.localStorageAvailable();

    if (hasLocalStorage) {
      try {
        this.apiKeys = JSON.parse(localStorage.getItem('remotestorage:api-keys')) || {};
      } catch(exc) {
        // ignored
      }
      this.setBackend(localStorage.getItem('remotestorage:backend') || 'remotestorage');
    }

    var origOn = this.on;

    this.on = function (eventName, handler) {
      if (eventName === 'ready' && this.remote.connected && this._allLoaded) {
        setTimeout(handler, 0);
      } else if (eventName === 'features-loaded' && this._allLoaded) {
        setTimeout(handler, 0);
      }
      return origOn.call(this, eventName, handler);
    };

    this._init();

    this.fireInitial = function () {
      if (this.local) {
        setTimeout(this.local.fireInitial.bind(this.local), 0);
      }
    }.bind(this);

    this.on('ready', this.fireInitial.bind(this));
  };

  RemoteStorage.SyncedGetPutDelete = SyncedGetPutDelete;

  RemoteStorage.DiscoveryError = function (message) {
    Error.apply(this, arguments);
    this.message = message;
  };

  RemoteStorage.DiscoveryError.prototype = Object.create(Error.prototype);

  RemoteStorage.Unauthorized = function () { Error.apply(this, arguments); };
  RemoteStorage.Unauthorized.prototype = Object.create(Error.prototype);

  /**
   * Method: RemoteStorage.log
   *
   * Log using console.log, when remoteStorage logging is enabled.
   *
   * You can enable logging with <enableLog>.
   *
   * (In node.js you can also enable logging during remoteStorage object
   * creation. See: <RemoteStorage>).
   */
  RemoteStorage.log = function () {
    if (RemoteStorage.config.logging) {
      console.log.apply(console, arguments);
    }
  };

  RemoteStorage.config = {
    logging: false,
    changeEvents: {
      local:    true,
      window:   false,
      remote:   true,
      conflict: true
    },
    discoveryTimeout: 10000,
    cordovaRedirectUri: undefined
  };

  RemoteStorage.prototype = {

    /**
     * Method: displayWidget
     *
     * Displays the widget at the top right of the page. Make sure to call this function
     * once on every pageload (after the html 'body' tag), unless you use a custom widget.
     *
     * Parameters:
     *
     *   domID: identifier of the DOM element which should embody the widget (optional)
     */
     // (see src/widget.js for implementation)

    /**
     * Property: remote
     *
     * Properties:
     *
     *   connected   - Boolean, whether or not a remote store is connected
     *   online      - Boolean, whether last sync action was successful or not
     *   userAddress - String, the user address of the connected user
     *   properties  - String, the properties of the WebFinger link
     */

    /**
     * Method: scope
     *
     * Returns a BaseClient with a certain scope (base path). Please use this method
     * only for debugging, and always use defineModule instead, to get access to a
     * BaseClient from a module in an app.
     *
     * Parameters:
     *
     *   scope - A string, with a leading and a trailing slash, specifying the
     *           base path of the BaseClient that will be returned.
     *
     * Code example:
     *
     * (start code)
     * remoteStorage.scope('/pictures/').getListing('');
     * remoteStorage.scope('/public/pictures/').getListing('');
     */

    /**
     * Method: startSync
     *
     * Start synchronization with remote storage, downloading and uploading any
     * changes within the cached paths.
     *
     * Please consider: local changes will attempt sync immediately, and remote
     * changes should also be synced timely when using library defaults. So
     * this is mostly useful for letting users sync manually, when pressing a
     * sync button for example. This might feel safer to them sometimes, esp.
     * when shifting between offline and online a lot.
     */
     // (see src/sync.js for implementation)

    /**
     * Method: connect
     *
     * Connect to a remoteStorage server.
     *
     * Parameters:
     *   userAddress        - The user address (user@host) to connect to.
     *   token              - (optional) A bearer token acquired beforehand
     *
     * Discovers the WebFinger profile of the given user address and initiates
     * the OAuth dance.
     *
     * This method must be called *after* all required access has been claimed.
     * When using the connect widget, it will call this method itself.
     *
     * Special cases:
     *
     * 1. If a bearer token is supplied as second argument, the OAuth dance
     *    will be skipped and the supplied token be used instead. This is
     *    useful outside of browser environments, where the token has been
     *    acquired in a different way.
     *
     * 2. If the Webfinger profile for the given user address doesn't contain
     *    an auth URL, the library will assume that client and server have
     *    established authorization among themselves, which will omit bearer
     *    tokens in all requests later on. This is useful for example when using
     *    Kerberos and similar protocols.
     */
    connect: function (userAddress, token) {
      this.setBackend('remotestorage');
      if (userAddress.indexOf('@') < 0) {
        this._emit('error', new RemoteStorage.DiscoveryError("User address doesn't contain an @."));
        return;
      }

      if (global.cordova) {
        if (typeof RemoteStorage.config.cordovaRedirectUri !== 'string') {
          this._emit('error', new RemoteStorage.DiscoveryError("Please supply a custom HTTPS redirect URI for your Cordova app"));
          return;
        }
        if (!global.cordova.InAppBrowser) {
          this._emit('error', new RemoteStorage.DiscoveryError("Please include the InAppBrowser Cordova plugin to enable OAuth"));
          return;
        }
      }

      this.remote.configure({
        userAddress: userAddress
      });
      this._emit('connecting');

      var discoveryTimeout = setTimeout(function () {
        this._emit('error', new RemoteStorage.DiscoveryError("No storage information found for this user address."));
      }.bind(this), RemoteStorage.config.discoveryTimeout);

      RemoteStorage.Discover(userAddress).then(function (info) {
        // Info contains fields: href, storageApi, authURL (optional), properties

        clearTimeout(discoveryTimeout);
        this._emit('authing');
        info.userAddress = userAddress;
        this.remote.configure(info);
        if (! this.remote.connected) {
          if (info.authURL) {
            if (typeof token === 'undefined') {
              // Normal authorization step; the default way to connect
              this.authorize(info.authURL, RemoteStorage.config.cordovaRedirectUri);
            } else if (typeof token === 'string') {
              // Token supplied directly by app/developer/user
              RemoteStorage.log('Skipping authorization sequence and connecting with known token');
              this.remote.configure({ token: token });
            } else {
              throw new Error("Supplied bearer token must be a string");
            }
          } else {
            // In lieu of an excplicit authURL, assume that the browser and
            // server handle any authorization needs; for instance, TLS may
            // trigger the browser to use a client certificate, or a 401 Not
            // Authorized response may make the browser send a Kerberos ticket
            // using the SPNEGO method.
            this.impliedauth();
          }
        }
      }.bind(this), function(err) {
        clearTimeout(discoveryTimeout);
        this._emit('error', new RemoteStorage.DiscoveryError("No storage information found for this user address."));
      }.bind(this));
    },

    /**
     * Method: disconnect
     *
     * "Disconnect" from remotestorage server to terminate current session.
     * This method clears all stored settings and deletes the entire local
     * cache.
     */
    disconnect: function () {
      if (this.remote) {
        this.remote.configure({
          userAddress: null,
          href: null,
          storageApi: null,
          token: null,
          properties: null
        });
      }
      this._setGPD({
        get: this._pendingGPD('get'),
        put: this._pendingGPD('put'),
        delete: this._pendingGPD('delete')
      });
      var n = this._cleanups.length, i = 0;

      var oneDone = function () {
        i++;
        if (i >= n) {
          this._init();
          RemoteStorage.log('Done cleaning up, emitting disconnected and disconnect events');
          this._emit('disconnected');
        }
      }.bind(this);

      if (n > 0) {
        this._cleanups.forEach(function (cleanup) {
          var cleanupResult = cleanup(this);
          if (typeof(cleanupResult) === 'object' && typeof(cleanupResult.then) === 'function') {
            cleanupResult.then(oneDone);
          } else {
            oneDone();
          }
        }.bind(this));
      } else {
        oneDone();
      }
    },

    setBackend: function (what) {
      this.backend = what;
      if (hasLocalStorage) {
        if (what) {
          localStorage.setItem('remotestorage:backend', what);
        } else {
          localStorage.removeItem('remotestorage:backend');
        }
      }
    },

    /**
     * Method: onChange
     *
     * Add a "change" event handler to the given path. Whenever a "change"
     * happens (as determined by the backend, such as e.g.
     * <RemoteStorage.IndexedDB>) and the affected path is equal to or below
     * the given 'path', the given handler is called.
     *
     * You should usually not use this method directly, but instead use the
     * "change" events provided by <RemoteStorage.BaseClient>.
     *
     * Parameters:
     *   path    - Absolute path to attach handler to.
     *   handler - Handler function.
     */
    onChange: function (path, handler) {
      if (! this._pathHandlers.change[path]) {
        this._pathHandlers.change[path] = [];
      }
      this._pathHandlers.change[path].push(handler);
    },

    /**
     * Method: enableLog
     *
     * Enable remoteStorage logging.
     */
    enableLog: function () {
      RemoteStorage.config.logging = true;
    },

    /**
     * Method: disableLog
     *
     * Disable remoteStorage logging
     */
    disableLog: function () {
      RemoteStorage.config.logging = false;
    },

    /**
     * Method: log
     *
     * The same as <RemoteStorage.log>.
     */
    log: function () {
      RemoteStorage.log.apply(RemoteStorage, arguments);
    },

    /**
     * Method: setApiKeys (experimental)
     *
     * Set API keys for (currently) GoogleDrive and/or Dropbox backend support.
     * See also the 'backends' example in the starter-kit. Note that support for
     * both these backends is still experimental.
     *
     * Parameters:
     *   type - string, either 'googledrive' or 'dropbox'
     *   keys - object, with one string field; 'clientId' for GoogleDrive, or
     *          'appKey' for Dropbox.
     *
     */
    setApiKeys: function (type, keys) {
      if (keys) {
        this.apiKeys[type] = keys;
        if (type === 'dropbox' && (typeof this.dropbox === 'undefined' ||
                                   this.dropbox.clientId !== keys.appKey)) {
          RemoteStorage.Dropbox._rs_init(this);
        } else if (type === 'googledrive' && (typeof this.googledrive === 'undefined' ||
                                              this.googledrive.clientId !== keys.clientId)) {
          RemoteStorage.GoogleDrive._rs_init(this);
        }
      } else {
        delete this.apiKeys[type];
      }
      if (hasLocalStorage) {
        localStorage.setItem('remotestorage:api-keys', JSON.stringify(this.apiKeys));
      }
    },

    /**
     * Method: setCordovaRedirectUri
     *
     * Set redirect URI to be used for the OAuth redirect within the
     * in-app-browser window in Cordova apps.
     *
     * Parameters:
     *   uri - string, valid HTTP(S) URI
     */
    setCordovaRedirectUri: function (uri) {
      if (typeof uri !== 'string' || !uri.match(/http(s)?\:\/\//)) {
        throw new Error("Cordova redirect URI must be a URI string");
      }
      RemoteStorage.config.cordovaRedirectUri = uri;
    },

    /**
     ** INITIALIZATION
     **/

    _init: function () {
      var self = this,
          readyFired = false;

      function fireReady() {
        try {
          if (!readyFired) {
            self._emit('ready');
            readyFired = true;
          }
        } catch(e) {
          console.error("'ready' failed: ", e, e.stack);
          self._emit('error', e);
        }
      }

      this._loadFeatures(function (features) {
        this.log('[RemoteStorage] All features loaded');
        this.local = features.local && new features.local();
        // this.remote set by WireClient._rs_init as lazy property on
        // RS.prototype

        if (this.local && this.remote) {
          this._setGPD(SyncedGetPutDelete, this);
          this._bindChange(this.local);
        } else if (this.remote) {
          this._setGPD(this.remote, this.remote);
        }

        if (this.remote) {
          this.remote.on('connected', function (){
            fireReady();
            self._emit('connected');
          });
          this.remote.on('not-connected', function (){
            fireReady();
            self._emit('not-connected');
          });
          if (this.remote.connected) {
            fireReady();
            self._emit('connected');
          }

          if (!this.hasFeature('Authorize')) {
            this.remote.stopWaitingForToken();
          }
        }

        this._collectCleanupFunctions();

        try {
          this._allLoaded = true;
          this._emit('features-loaded');
        } catch(exc) {
          logError(exc);
          this._emit('error', exc);
        }
        this._processPending();
      }.bind(this));
    },

    _collectCleanupFunctions: function () {
      this._cleanups = [];
      for (var i=0; i < this.features.length; i++) {
        var cleanup = this.features[i].cleanup;
        if (typeof(cleanup) === 'function') {
          this._cleanups.push(cleanup);
        }
      }
    },

    /**
     ** FEATURE DETECTION
     **/
    _loadFeatures: function (callback) {
      var featureList = [
        'WireClient',
        'I18n',
        'Dropbox',
        'GoogleDrive',
        'Access',
        'Caching',
        'Discover',
        'Authorize',
        'Widget',
        'IndexedDB',
        'LocalStorage',
        'InMemoryStorage',
        'Sync',
        'BaseClient',
        'Env'
      ];
      var features = [];
      var featuresDone = 0;
      var self = this;

      function featureDone() {
        featuresDone++;
        if (featuresDone === featureList.length) {
          setTimeout(function () {
            features.caching = !!RemoteStorage.Caching;
            features.sync = !!RemoteStorage.Sync;
            [
              'IndexedDB',
              'LocalStorage',
              'InMemoryStorage'
            ].some(function (cachingLayer) {
              if (features.some(function (feature) { return feature.name === cachingLayer; })) {
                features.local = RemoteStorage[cachingLayer];
                return true;
              }
            });
            self.features = features;
            callback(features);
          }, 0);
        }
      }

      function featureInitialized(name) {
        self.log("[RemoteStorage] [FEATURE "+name+"] initialized.");
        features.push({
          name : name,
          init :  RemoteStorage[name]._rs_init,
          supported : true,
          cleanup : RemoteStorage[name]._rs_cleanup
        });
        featureDone();
      }

      function featureFailed(name, err) {
        self.log("[RemoteStorage] [FEATURE "+name+"] initialization failed ( "+err+")");
        featureDone();
      }

      function featureSupported(name, success) {
        self.log("[RemoteStorage] [FEATURE "+name+"]" + success ? "":" not"+" supported");
        if (!success) {
          featureDone();
        }
      }

      function initFeature(name) {
        var initResult;
        try {
          initResult = RemoteStorage[name]._rs_init(self);
        } catch(e) {
          featureFailed(name, e);
          return;
        }
        if (typeof(initResult) === 'object' && typeof(initResult.then) === 'function') {
          initResult.then(
            function (){ featureInitialized(name); },
            function (err){ featureFailed(name, err); }
          );
        } else {
          featureInitialized(name);
        }
      }

      featureList.forEach(function (featureName) {
        self.log("[RemoteStorage] [FEATURE " + featureName + "] initializing...");
        var impl = RemoteStorage[featureName];
        var supported;

        if (impl) {
          supported = !impl._rs_supported || impl._rs_supported();

          if (typeof supported === 'object') {
            supported.then(
              function (){
                featureSupported(featureName, true);
                initFeature(featureName);
              },
              function (){
                featureSupported(featureName, false);
              }
            );
          } else if (typeof supported === 'boolean') {
            featureSupported(featureName, supported);
            if (supported) {
              initFeature(featureName);
            }
          }
        } else {
          featureSupported(featureName, false);
        }
      });
    },

    /**
     * Method: hasFeature
     *
     * Checks whether a feature is enabled or not within remoteStorage.
     * Returns a boolean.
     *
     * Parameters:
     *   name - Capitalized name of the feature. e.g. Authorize, or IndexedDB
     *
     * Example:
     *   (start code)
     *   if (remoteStorage.hasFeature('LocalStorage')) {
     *     console.log('LocalStorage is enabled!');
     *   }
     *   (end code)
     *
     */
    hasFeature: function (feature) {
      for (var i = this.features.length - 1; i >= 0; i--) {
        if (this.features[i].name === feature) {
          return this.features[i].supported;
        }
      }
      return false;
    },

    /**
     ** GET/PUT/DELETE INTERFACE HELPERS
     **/

    _setGPD: function (impl, context) {
      function wrap(func) {
        return function () {
          return func.apply(context, arguments)
            .then(emitUnauthorized.bind(this));
        };
      }
      this.get = wrap(impl.get);
      this.put = wrap(impl.put);
      this.delete = wrap(impl.delete);
    },

    _pendingGPD: function (methodName) {
      return function () {
        var pending = Promise.defer();
        this._pending.push({
          method: methodName,
          args: Array.prototype.slice.call(arguments),
          promise: pending
        });
        return pending.promise;
      }.bind(this);
    },

    _processPending: function () {
      this._pending.forEach(function (pending) {
        try {
          this[pending.method].apply(this, pending.args).then(pending.promise.resolve, pending.promise.reject);
        } catch(e) {
          pending.promise.reject(e);
        }
      }.bind(this));
      this._pending = [];
    },

    /**
     ** CHANGE EVENT HANDLING
     **/

    _bindChange: function (object) {
      object.on('change', this._dispatchEvent.bind(this, 'change'));
    },

    _dispatchEvent: function (eventName, event) {
      var self = this;
      Object.keys(this._pathHandlers[eventName]).forEach(function (path) {
        var pl = path.length;
        if (event.path.substr(0, pl) === path) {
          self._pathHandlers[eventName][path].forEach(function (handler) {
            var ev = {};
            for (var key in event) { ev[key] = event[key]; }
            ev.relativePath = event.path.replace(new RegExp('^' + path), '');
            try {
              handler(ev);
            } catch(e) {
              console.error("'change' handler failed: ", e, e.stack);
              self._emit('error', e);
            }
          });
        }
      });
    }
  };

  /**
   * Property: connected
   *
   * Boolean property indicating if remoteStorage is currently connected.
   */
  Object.defineProperty(RemoteStorage.prototype, 'connected', {
    get: function () {
      return this.remote.connected;
    }
  });

  /**
   * Property: access
   *
   * Tracking claimed access scopes. A <RemoteStorage.Access> instance.
   *
   *
   * Property: caching
   *
   * Caching settings. A <RemoteStorage.Caching> instance.
   *
   * Not available in no-cache builds.
   *
   *
   * Property: remote
   *
   * Access to the remote backend used. Usually a <RemoteStorage.WireClient>.
   *
   *
   * Property: local
   *
   * Access to the local caching backend used. Usually either a
   * <RemoteStorage.IndexedDB> or <RemoteStorage.LocalStorage> instance.
   *
   * Not available in no-cache builds.
   */

  global.RemoteStorage = RemoteStorage;

  if ((typeof module === 'object') && (typeof module.exports !== 'undefined')){
    module.exports = RemoteStorage;
  }
})(typeof(window) !== 'undefined' ? window : global);
